import time
import numpy as np
import math
import torch
from torch import Tensor
from eindex import eindex
from typing import Optional, Union
import torch.nn.functional as F
import einops
from jaxtyping import Float, Int
from rich import print as rprint
from rich.table import Table
from collections import defaultdict
from transformer_lens import utils, HookedTransformer
from tqdm.auto import tqdm

Arr = np.ndarray

from sae_vis.model_fns import (
    AutoEncoder,
    TransformerLensWrapper,
    to_resid_dir,
)
from sae_vis.utils_fns import (
    k_largest_indices,
    random_range_indices,
    QuantileCalculator,
    TopK,
    get_device,
    BatchedCorrCoef,
)
from sae_vis.data_storing_fns import (
    SequenceData,
    SequenceGroupData,
    SequenceMultiGroupData,
    FeatureTablesData,
    FeatureData,
    SaeVisData,
    ActsHistogramData,
    LogitsHistogramData,
    LogitsTableData,
)
from sae_vis.data_config_classes import (
    SaeVisLayoutConfig,
    SaeVisConfig,
    SequencesConfig,
)

device = get_device()





def compute_feat_acts(
    model_acts: Float[Tensor, "batch seq d_in"],
    feature_idx: Int[Tensor, "feats"],
    encoder: AutoEncoder,
    encoder_B: Optional[AutoEncoder] = None,
    corrcoef_neurons: Optional[BatchedCorrCoef] = None,
    corrcoef_encoder_B: Optional[BatchedCorrCoef] = None,
) -> Float[Tensor, "batch seq feats"]:
    '''
    This function computes the feature activations, given a bunch of model data. It also updates the rolling correlation
    coefficient objects, if they're given.

    Args:
        model_acts: Float[Tensor, "batch seq d_in"]
            The activations of the model, which the SAE was trained on.
        feature_idx: Int[Tensor, "feats"]
            The features we're computing the activations for. This will be used to index the encoder's weights.
        encoder: AutoEncoder
            The encoder object, which we use to calculate the feature activations.
        encoder_B: Optional[AutoEncoder]
            The encoder-B object, which we use to calculate the feature activations.
        corrcoef_neurons: Optional[BatchedCorrCoef]
            The object storing the minimal data necessary to compute corrcoef between feature activations & neurons.
        corrcoef_encoder_B: Optional[BatchedCorrCoef]
            The object storing minimal data to compute corrcoef between feature activations & encoder-B features.
    '''
    # Get the feature act direction by indexing encoder.W_enc, and the bias by indexing encoder.b_enc
    feature_act_dir = encoder.W_enc[:, feature_idx] # (d_in, feats)
    feature_bias = encoder.b_enc[feature_idx] # (feats,)

    # Calculate & store feature activations (we need to store them so we can get the sequence & histogram vis later)
    x_cent = model_acts - encoder.b_dec
    feat_acts_pre = einops.einsum(x_cent, feature_act_dir, "batch seq d_in, d_in feats -> batch seq feats")
    feat_acts = F.relu(feat_acts_pre + feature_bias)

    # Update the CorrCoef object between feature activation & neurons
    if corrcoef_neurons is not None:
        corrcoef_neurons.update(
            einops.rearrange(feat_acts, "batch seq feats -> feats (batch seq)"),
            einops.rearrange(model_acts, "batch seq d_in -> d_in (batch seq)"),
        )
        
    # Calculate encoder-B feature acts (we don't need to store them, cause it's just for the left-side feature tables)
    if corrcoef_encoder_B is not None:
        assert encoder_B is not None,\
            "Error: you need to supply an encoder-B object if you want to calculate encoder-B feature activations."
        x_cent_B = model_acts - encoder_B.b_dec
        feat_acts_pre_B = einops.einsum(x_cent_B, encoder_B.W_enc, "batch seq d_in, d_in d_hidden -> batch seq d_hidden")
        feat_acts_B = F.relu(feat_acts_pre_B + encoder.b_enc)

        # Update the CorrCoef object between feature activation & encoder-B features
        corrcoef_encoder_B.update(
            einops.rearrange(feat_acts, "batch seq feats -> feats (batch seq)"),
            einops.rearrange(feat_acts_B, "batch seq d_hidden -> d_hidden (batch seq)"),
        )

    return feat_acts



@torch.inference_mode()
def parse_feature_data(
    tokens: Int[Tensor, "batch seq"],
    feature_indices: Union[int, list[int]],
    all_feat_acts: Float[Tensor, "... feats"],
    feature_resid_dir: Float[Tensor, "feats d_model"],
    all_resid_post: Float[Tensor, "... d_model"],
    W_U: Float[Tensor, "d_model d_vocab"],
    cfg: SaeVisConfig,
    feature_out_dir: Optional[Float[Tensor, "feats d_out"]] = None,
    corrcoef_neurons: Optional[BatchedCorrCoef] = None,
    corrcoef_encoder_B: Optional[BatchedCorrCoef] = None,
    progress: Optional[list[tqdm]] = None,
) -> tuple[SaeVisData, dict[str, float]]:
    """Convert generic activation data into a SaeVisData object, which can be used to create the feature-centric vis.

    This function exists so that feature dashboards can be generated without using our AutoEncoder or
    TransformerLens(Wrapper) classes. We pass through W_U & other data needed for computing the logit lens, so we don't
    need the models.

    Args:
        tokens: Int[Tensor, "batch seq"]
            The tokens we'll be using to get the feature activations.
        
        feature_indices: Union[int, list[int]]
            The features we're actually computing. These might just be a subset of the model's full features.
        
        all_feat_acts: Float[Tensor, "... feats"]
            The activations values of the features across the batch & sequence.
        
        feature_resid_dir: Float[Tensor, "feats d_model"]
            The directions that each feature writes to the residual stream.
            For example, feature_resid_dir = encoder.W_dec[feature_indices] # [feats d_autoencoder]
        
        all_resid_post: Float[Tensor, "... d_model"]
            The activations of the final layer of the model before the unembed. 
        
        W_U: Float[Tensor, "d_model d_vocab"]
            The model's unembed weights for the logit lens.
        
        cfg: SaeVisConfig
            Feature visualization parameters, containing a bunch of other stuff. See the SaeVisConfig docstring for
            more information.
        
        feature_out_dir: Optional[Float[Tensor, "feats d_out"]]
            The directions that each SAE feature writes to the residual stream. This will be the same as 
            feature_resid_dir if the SAE is in the residual stream (as we will assume if it not provided)
            For example, feature_out_dir = encoder.W_dec[feature_indices] # [feats d_autoencoder]
        
        corrcoef_neurons: Optional[BatchedCorrCoef]
            The object storing the minimal data necessary to compute corrcoef between feature activations & neurons.

        corrcoef_encoder_B: Optional[BatchedCorrCoef]
            The object storing minimal data to compute corrcoef between feature activations & encoder-B features.

        progress: Optional[list[tqdm]]
            An optional list containing progress bars for the forward passes and the sequence data. This is used to
            update the progress bars as the computation runs.
    
    Returns:
        sae_vis_data: SaeVisData
            Containing data for creating each feature visualization, as well as data for rank-ordering the feature
            visualizations when it comes time to make the prompt-centric view (the `feature_act_quantiles` attribute).

        time_logs: dict[str, float]
            A dictionary containing the time taken for each step of the computation. This is optionally printed at the
            end of the `get_feature_data` function, if `cfg.verbose` is set to True.
    """
    time_logs = {
        "(4) Getting data for tables": 0.0,
        "(5) Getting data for histograms": 0.0,
        "(6) Getting data for sequences": 0.0,
        "(7) Getting data for quantiles": 0.0,
    }
    t0 = time.time()
    
    # Make feature_indices a list, for convenience
    if isinstance(feature_indices, int): feature_indices = [feature_indices]

    assert feature_resid_dir.shape[0] == len(feature_indices),\
        f"Num features in feature_resid_dir ({feature_resid_dir.shape[0]}) doesn't match {len(feature_indices)=}"
    
    if feature_out_dir is not None:
        assert feature_out_dir.shape[0] == len(feature_indices),\
            f"Num features in feature_out_dir ({feature_resid_dir.shape[0]}) doesn't match {len(feature_indices)=}"

    # ! Data setup code (defining the main objects we'll eventually return)
    feature_data_dict: dict[int, FeatureData] = {feat: FeatureData() for feat in feature_indices}

    # We're using `cfg.feature_centric_layout` to figure out what data we'll need to calculate during this function
    layout = cfg.feature_centric_layout
    assert isinstance(layout, SaeVisLayoutConfig),\
        f"Error: cfg.feature_centric_layout must be a SaeVisLayoutConfig object, got {type(layout)}"


    # ! Calculate all data for the left-hand column visualisations, i.e. the 3 tables

    if layout.feat_tables_cfg is not None and feature_out_dir is not None:

        # Table 1: neuron alignment, based on decoder weights
        top3_neurons_aligned = TopK(tensor=feature_out_dir, k=layout.feat_tables_cfg.n_rows, largest=True)
        feature_out_l1_norm = feature_out_dir.abs().sum(dim=-1, keepdim=True)
        pct_of_l1: Arr = np.absolute(top3_neurons_aligned.values) / utils.to_numpy(feature_out_l1_norm)
        
        # Table 2: neurons correlated with this feature, based on their activations
        assert isinstance(corrcoef_neurons, BatchedCorrCoef),\
            f"Error: corrcoef_neurons should be of type BatchedCorrCoef, instead we got {type(corrcoef_neurons)}"
        neuron_indices, neuron_pearson, neuron_cossim = corrcoef_neurons.topk_pearson(k=layout.feat_tables_cfg.n_rows)
        
        # Table 3: encoder-B features correlated with this feature, based on their activations. Note that we don't need
        # to calculate this if we're not using encoder_B
        encB_indices = encB_pearson = encB_cossim = [[] for _ in feature_indices]
        if (corrcoef_encoder_B is not None):
            assert isinstance(corrcoef_encoder_B, BatchedCorrCoef),\
                f"Error: corrcoef_encoder_B should be of type BatchedCorrCoef, instead we got {type(corrcoef_encoder_B)}"
            encB_indices, encB_pearson, encB_cossim = corrcoef_encoder_B.topk_pearson(k=layout.feat_tables_cfg.n_rows)
        
        # Add all this data to the list of FeatureTablesData objects
        for i, feat in enumerate(feature_indices):
            feature_data_dict[feat].feature_tables_data = FeatureTablesData(
                neuron_alignment_indices = top3_neurons_aligned[i].indices.tolist(),
                neuron_alignment_values = top3_neurons_aligned[i].values.tolist(),
                neuron_alignment_l1 = pct_of_l1[i].tolist(),
                correlated_neurons_indices = neuron_indices[i],
                correlated_neurons_pearson = neuron_pearson[i],
                correlated_neurons_cossim = neuron_cossim[i],
                correlated_features_indices = encB_indices[i],
                correlated_features_pearson = encB_pearson[i],
                correlated_features_cossim = encB_cossim[i],
            )

    time_logs["(4) Getting data for tables"] = time.time() - t0
    t0 = time.time()


    # ! Get all data for the middle column visualisations, i.e. the two histograms & the logit table

    # Get the logits of all features (i.e. the directions this feature writes to the logit output)
    logits = einops.einsum(feature_resid_dir, W_U, "feats d_model, d_model d_vocab -> feats d_vocab")
    if any(x is not None for x in [layout.act_hist_cfg, layout.logits_hist_cfg, layout.logits_table_cfg]):

        for i, (feat, logit_vector) in enumerate(zip(feature_indices, logits)):

            # Get logits histogram data (no title)
            if layout.logits_hist_cfg is not None:
                feature_data_dict[feat].logits_histogram_data = LogitsHistogramData.from_data(
                    data = logit_vector,
                    n_bins = layout.logits_hist_cfg.n_bins,
                    tickmode = '5 ticks',
                    title = None,
                )

            # Get data for feature activations histogram (including the title!)
            if layout.act_hist_cfg is not None:
                feat_acts = all_feat_acts[..., i]
                nonzero_feat_acts = feat_acts[feat_acts > 0]
                frac_nonzero = nonzero_feat_acts.numel() / feat_acts.numel()
                feature_data_dict[feat].acts_histogram_data = ActsHistogramData.from_data(
                    data = nonzero_feat_acts,
                    n_bins = layout.act_hist_cfg.n_bins,
                    tickmode = '5 ticks',
                    title = f"ACTIVATIONS<br>DENSITY = {frac_nonzero:.3%}"
                )

            if layout.logits_table_cfg is not None:
                # Get logits table data
                top_logits = TopK(logit_vector, k=layout.logits_table_cfg.n_rows, largest=True)
                bottom_logits = TopK(logit_vector, k=layout.logits_table_cfg.n_rows, largest=False)

                top_logits, top_token_ids = top_logits.values.tolist(), top_logits.indices.tolist()
                bottom_logits, bottom_token_ids = bottom_logits.values.tolist(), bottom_logits.indices.tolist()
                
                # Create a MiddlePlotsData object from this, and add it to the dict
                feature_data_dict[feat].logits_table_data = LogitsTableData(
                    bottom_logits = bottom_logits,
                    bottom_token_ids = bottom_token_ids,
                    top_logits = top_logits,
                    top_token_ids = top_token_ids,
                )

    time_logs["(5) Getting data for histograms"] = time.time() - t0
    t0 = time.time()


    # ! Calculate all data for the right-hand visualisations, i.e. the sequences

    if layout.seq_cfg is not None:
        for i, feat in enumerate(feature_indices):

            # Add this feature's sequence data to the list
            feature_data_dict[feat].sequence_data = get_sequences_data(
                tokens = tokens,
                feat_acts = all_feat_acts[..., i],
                feat_logits = logits[i],
                resid_post = all_resid_post,
                feature_resid_dir = feature_resid_dir[i],
                W_U = W_U,
                seq_cfg = layout.seq_cfg,
            )
            # Update the 2nd progress bar (fwd passes & getting sequence data dominates the runtime of these computations)
            if progress is not None:
                progress[1].update(1)

    time_logs["(6) Getting data for sequences"] = time.time() - t0
    t0 = time.time()


    # ! Get the quantiles, which will be useful for the prompt-centric visualisation
    feature_act_quantiles = QuantileCalculator.from_data(data=einops.rearrange(all_feat_acts, "b s feats -> feats (b s)"))
    time_logs["(7) Getting data for quantiles"] = time.time() - t0
    t0 = time.time()


    # ! Return the output, as a dict of FeatureData items
    sae_vis_data = SaeVisData(feature_data_dict, feature_act_quantiles, cfg)
    return sae_vis_data, time_logs




@torch.inference_mode()
def _get_feature_data(
    encoder: AutoEncoder,
    encoder_B: Optional[AutoEncoder],
    model: TransformerLensWrapper,
    tokens: Int[Tensor, "batch seq"],
    feature_indices: Union[int, list[int]],
    cfg: SaeVisConfig,
    progress: Optional[list[tqdm]] = None,
) -> tuple[SaeVisData, dict[str, float]]:
    '''
    Gets data that will be used to create the sequences in the feature-centric HTML visualisation.
    
    Note - this function isn't called directly by the user, it actually gets called by the `get_feature_data` function
    which does exactly the same thing except it also batches this computation by features (in accordance with the
    arguments `features` and `minibatch_size_features` from the SaeVisConfig object).

    Args:
        encoder: AutoEncoder
            The encoder whose features we'll be analyzing.

        encoder_B: AutoEncoder
            The encoder we'll be using as a reference (i.e. finding the B-features with the highest correlation). This
            is only necessary if we're generating the left-hand tables (i.e. cfg.include_left_tables=True).

        model: TransformerLensWrapper
            The model we'll be using to get the feature activations. It's a wrapping of the base TransformerLens model.

        tokens: Int[Tensor, "batch seq"]
            The tokens we'll be using to get the feature activations.

        feature_indices: Union[int, list[int]]
            The features we're actually computing. These might just be a subset of the model's full features.

        cfg: SaeVisConfig
            Feature visualization parameters, containing a bunch of other stuff. See the SaeVisConfig docstring for
            more information.

        progress: Optional[list[tqdm]]
            An optional list containing progress bars for the forward passes and the sequence data. This is used to
            update the progress bars as the computation runs.

    Returns:
        sae_vis_data: SaeVisData
            Containing data for creating each feature visualization, as well as data for rank-ordering the feature
            visualizations when it comes time to make the prompt-centric view (the `feature_act_quantiles` attribute).

        time_log: dict[str, float]
            A dictionary containing the time taken for each step of the computation. This is optionally printed at the
            end of the `get_feature_data` function, if `cfg.verbose` is set to True.
    '''
    # ! Boring setup code

    time_logs = {
        "(1) Initialization": 0.0,
        "(2) Forward passes to gather model activations": 0.0,
        "(3) Computing feature acts from model acts": 0.0,
    }

    t0 = time.time()

    # Make feature_indices a list, for convenience
    if isinstance(feature_indices, int): feature_indices = [feature_indices]
    feature_idx_tensor = torch.tensor(feature_indices, device=device)

    # Get tokens into minibatches, for the fwd pass
    token_minibatches = (tokens,) if cfg.minibatch_size_tokens is None else tokens.split(cfg.minibatch_size_tokens)
    token_minibatches = [tok.to(device) for tok in token_minibatches]

    # ! Data setup code (defining the main objects we'll eventually return, for each of 5 possible vis components)

    # Create lists to store the feature activations & final values of the residual stream
    all_resid_post = []
    all_feat_acts = []

    # Create objects to store the data for computing correlation coefficients (for left tables), if we're using them
    corrcoef_neurons = BatchedCorrCoef()
    corrcoef_encoder_B = BatchedCorrCoef() if encoder_B is not None else None

    # Get encoder & decoder directions
    feature_out_dir = encoder.W_dec[feature_indices] # [feats d_autoencoder]
    feature_resid_dir = to_resid_dir(feature_out_dir, model) # [feats d_model]

    time_logs["(1) Initialization"] = time.time() - t0

    # ! Compute & concatenate together all feature activations & post-activation function values

    for minibatch in token_minibatches:

        # Fwd pass, get model activations
        t0 = time.time()
        residual, model_acts = model.forward(minibatch, return_logits=False)
        time_logs["(2) Forward passes to gather model activations"] += time.time() - t0
        
        # Compute feature activations from this
        t0 = time.time()
        feat_acts = compute_feat_acts(model_acts, feature_idx_tensor, encoder, encoder_B, corrcoef_neurons, corrcoef_encoder_B)
        time_logs["(3) Computing feature acts from model acts"] += time.time() - t0

        # Add these to the lists (we'll eventually concat)
        all_feat_acts.append(feat_acts)
        all_resid_post.append(residual)

        # Update the 1st progress bar (fwd passes & getting sequence data dominates the runtime of these computations)
        if progress is not None:
            progress[0].update(1)

    all_feat_acts = torch.cat(all_feat_acts, dim=0)
    all_resid_post = torch.cat(all_resid_post, dim=0)

    # ! Use the data we've collected to make a MultiFeatureData object
    sae_vis_data, _time_logs = parse_feature_data(
        tokens = tokens,
        feature_indices = feature_indices,
        all_feat_acts = all_feat_acts,
        feature_resid_dir = feature_resid_dir,
        all_resid_post = all_resid_post,
        W_U = model.W_U,
        cfg = cfg,
        feature_out_dir = feature_out_dir,
        corrcoef_neurons = corrcoef_neurons,
        corrcoef_encoder_B = corrcoef_encoder_B,
        progress = progress,
    )

    assert set(time_logs.keys()) & set(_time_logs.keys()) == set(),\
        f"Invalid keys: {set(time_logs.keys()) & set(_time_logs.keys())} should have zero overlap"

    time_logs.update(_time_logs)

    return sae_vis_data, time_logs




@torch.inference_mode()
def get_feature_data(
    encoder: AutoEncoder,
    model: HookedTransformer,
    tokens: Int[Tensor, "batch seq"],
    cfg: SaeVisConfig,
    encoder_B: Optional[AutoEncoder] = None,
) -> SaeVisData:
    '''
    This is the main function which users will run to generate the feature visualization data. It batches this
    computation over features, in accordance with the arguments in the SaeVisConfig object (we don't want to compute all
    the features at once, since might give OOMs).

    See the `_get_feature_data` function for an explanation of the arguments, as well as a more detailed explanation
    of what this function is doing.

    The return object is the merged SaeVisData objects returned by the `_get_feature_data` function.
    '''
    # Apply random seed
    if cfg.seed is not None:
        torch.manual_seed(cfg.seed)
        np.random.seed(cfg.seed)

    # Create objects to store all the data we'll get from `_get_feature_data`
    sae_vis_data = SaeVisData()
    time_logs = defaultdict(float)

    # Slice tokens, if we're only doing a subset of them
    if cfg.batch_size is None:
        tokens = tokens
    else:
        tokens = tokens[:cfg.batch_size]

    # Get a feature list (need to deal with the case where `cfg.features` is an int, or None)
    if cfg.features is None:
        features_list = list(range(encoder.cfg.d_hidden))
    elif isinstance(cfg.features, int):
        features_list = [cfg.features]
    else:
        features_list = list(cfg.features)

    # Break up the features into batches
    feature_batches = [x.tolist() for x in torch.tensor(features_list).split(cfg.minibatch_size_features)]
    # Calculate how many minibatches of tokens there will be (for the progress bar)
    n_token_batches = 1 if (cfg.minibatch_size_tokens is None) else math.ceil(len(tokens) / cfg.minibatch_size_tokens)
    # Get the denominator for each of the 2 progress bars
    totals = (n_token_batches*len(feature_batches), len(features_list))

    # Optionally add two progress bars (one for the forward passes, one for getting the sequence data)
    if cfg.verbose:
        progress = [
            tqdm(total=totals[0], desc="Forward passes to cache data for vis"),
            tqdm(total=totals[1], desc="Extracting vis data from cached data"),
        ]
    else:
        progress = None

    # If the model is from TransformerLens, we need to apply a wrapper to it for standardization
    assert isinstance(model, HookedTransformer), "Error: non-HookedTransformer models are not yet supported."
    assert isinstance(cfg.hook_point, str), "Error: cfg.hook_point must be a string"
    model_wrapper = TransformerLensWrapper(model, cfg.hook_point)

    # For each feat: get new data and update global data storage objects
    for features in feature_batches:
        new_feature_data, new_time_logs = _get_feature_data(
            encoder, encoder_B, model_wrapper, tokens, features, cfg, progress
        )
        sae_vis_data.update(new_feature_data)
        for key, value in new_time_logs.items():
            time_logs[key] += value

    # Now exited, make sure the progress bar is at 100%
    if progress is not None:
        for pbar in progress: pbar.n = pbar.total

    # If verbose, then print the output
    if cfg.verbose:
        total_time = sum(time_logs.values())
        table = Table("Task", "Time", "Pct %")
        for task, duration in time_logs.items():
            table.add_row(task, f"{duration:.2f}s", f"{duration/total_time:.1%}")
        rprint(table)

    return sae_vis_data




@torch.inference_mode()
def get_sequences_data(
    tokens: Int[Tensor, "batch seq"],
    feat_acts: Float[Tensor, "batch seq"],
    feat_logits: Float[Tensor, "d_vocab"],
    resid_post: Float[Tensor, "batch seq d_model"],
    feature_resid_dir: Float[Tensor, "d_model"],
    W_U: Float[Tensor, "d_model d_vocab"],
    seq_cfg: SequencesConfig,
) -> SequenceMultiGroupData:
    '''
    This function returns the data which is used to create the sequence visualizations (i.e. the right-hand column of
    the visualization). This is a multi-step process (the 4 steps are annotated in the code):

        (1) Find all the token groups (i.e. topk, bottomk, and quantile groups of activations). These are bold tokens.
        (2) Get the indices of all tokens we'll need data from, which includes a buffer around each bold token.
        (3) Extract the token IDs, feature activations & residual stream values for those positions
        (4) Compute the logit effect if this feature is ablated
            (4A) Use this to compute the most affected tokens by this feature (i.e. the vis hoverdata)
            (4B) Use this to compute the loss effect if this feature is ablated (i.e. the blue/red underlining)
        (5) Return all this data as a SequenceMultiGroupData object

    Args:
        tokens:
            The tokens we'll be extracting sequence data from.
        feat_acts:
            The activations of the feature we're interested in, for each token in the batch.
        feat_logits:
            The logit vector for this feature (used to generate histogram, and is needed here for the line-on-hover).
        resid_post:
            The residual stream values before final layernorm, for each token in the batch.
        feature_resid_dir:
            The direction this feature writes to the logit output (i.e. the direction we'll be erasing from resid_post).
        W_U:
            The model's unembedding matrix, which we'll use to get the logits.
        cfg:
            Feature visualization parameters, containing some important params e.g. num sequences per group.

    Returns:
        SequenceMultiGroupData
            This is a dataclass which contains a dict of SequenceGroupData objects, where each SequenceGroupData object
            contains the data for a particular group of sequences (i.e. the top-k, bottom-k, and the quantile groups).
    '''

    # ! (1) Find the tokens from each group

    # Get buffer, s.t. we're looking for bold tokens in the range `buffer[0] : buffer[1]`. For each bold token, we need
    # to see `seq_cfg.buffer[0]+1` behind it (plus 1 because we need the prev token to compute loss effect), and we need
    # to see `seq_cfg.buffer[1]` ahead of it
    buffer = (seq_cfg.buffer[0]+1, -seq_cfg.buffer[1])

    # Get the top-activating tokens
    indices = k_largest_indices(feat_acts, k=seq_cfg.top_acts_group_size, buffer=buffer)
    indices_dict = {f"TOP ACTIVATIONS<br>MAX = {feat_acts.max():.3f}": indices}

    # Get all possible indices. Note, we need to be able to look 1 back (feature activation on prev token is needed for
    # computing loss effect on this token)
    if seq_cfg.n_quantiles > 0:
        quantiles = torch.linspace(0, feat_acts.max().item(), seq_cfg.n_quantiles+1)
        for i in range(seq_cfg.n_quantiles-1, -1, -1):
            lower, upper = quantiles[i:i+2].tolist()
            pct = ((feat_acts >= lower) & (feat_acts <= upper)).float().mean()
            indices = random_range_indices(feat_acts, k=seq_cfg.quantile_group_size, bounds=(lower, upper), buffer=buffer)
            indices_dict[f"INTERVAL {lower:.3f} - {upper:.3f}<br>CONTAINS {pct:.3%}"] = indices

    # Concat all the indices together (in the next steps we do all groups at once)
    indices_bold = torch.concat(list(indices_dict.values())).cpu() # shape [batch 2]


    # ! (2) Get the buffer indices

    # Get the buffer indices, by adding a broadcasted arange object. At this point, indices_buf contains 1 more token
    # than the length of the sequences we'll see (because it also contains the token before the sequence starts).
    buffer_tensor = torch.arange(-seq_cfg.buffer[0] - 1, seq_cfg.buffer[1] + 1, device=indices_bold.device)
    indices_buf = einops.repeat(indices_bold, "batch two -> batch seq two", seq=seq_cfg.buffer[0] + seq_cfg.buffer[1] + 2)
    indices_buf = torch.stack([indices_buf[..., 0], indices_buf[..., 1] + buffer_tensor], dim=-1)

    
    # ! (3) Extract the token IDs, feature activations & residual stream values for those positions

    # Get the tokens which will be in our sequences
    token_ids = eindex(tokens, indices_buf[:, 1:], "[batch seq 0] [batch seq 1]") # shape [batch buf]

    # Now, we split into cases depending on whether we're computing the buffer or not. One kinda weird thing: we get
    # feature acts for 2 different reasons (token coloring & ablation), and in the case where we're computing the buffer
    # we need [:, 1:] for coloring and [:, :-1] for ablation, but when we're not we only need [:, bold] for both. So
    # we split on cases here.
    if seq_cfg.compute_buffer:
        feat_acts_buf = eindex(feat_acts, indices_buf, "[batch buf_plus1 0] [batch buf_plus1 1]")
        feat_acts_pre_ablation = feat_acts_buf[:, :-1]
        feat_acts_coloring = feat_acts_buf[:, 1:]
        resid_post_pre_ablation = eindex(resid_post, indices_buf[:, :-1], "[batch buf 0] [batch buf 1] d_model")
        # The tokens we'll use to index correct logits are the same as the ones which will be in our sequence
        correct_tokens = token_ids
    else:
        feat_acts_pre_ablation = eindex(feat_acts, indices_bold, "[batch 0] [batch 1]").unsqueeze(1)
        feat_acts_coloring = feat_acts_pre_ablation
        resid_post_pre_ablation = eindex(resid_post, indices_bold, "[batch 0] [batch 1] d_model").unsqueeze(1)
        # The tokens we'll use to index correct logits are the ones after bold
        indices_bold_next = torch.stack([indices_bold[:, 0], indices_bold[:, 1] + 1], dim=-1)
        correct_tokens = eindex(tokens, indices_bold_next, "[batch 0] [batch 1]").unsqueeze(1)


    # ! (4) Compute the logit effect if this feature is ablated

    # Get this feature's output vector, using an outer product over the feature activations for all tokens
    resid_post_feature_effect = feat_acts_pre_ablation[..., None] * feature_resid_dir # shape [batch buf d_model]

    # Do the ablations, and get difference in logprobs
    new_resid_post = resid_post_pre_ablation - resid_post_feature_effect
    new_logits = (new_resid_post / new_resid_post.std(dim=-1, keepdim=True)) @ W_U
    orig_logits = (resid_post_pre_ablation / resid_post_pre_ablation.std(dim=-1, keepdim=True)) @ W_U
    contribution_to_logprobs = orig_logits.log_softmax(dim=-1) - new_logits.log_softmax(dim=-1)

    # ! (4A) Use this to compute the most affected tokens by this feature
    # The TopK function can improve efficiency by masking the features which are zero
    acts_nonzero = feat_acts_pre_ablation.abs() > 1e-5 # shape [batch buf]
    top_contribution_to_logits = TopK(contribution_to_logprobs, k=seq_cfg.top_logits_hoverdata, largest=True, tensor_mask=acts_nonzero)
    bottom_contribution_to_logits = TopK(contribution_to_logprobs, k=seq_cfg.top_logits_hoverdata, largest=False, tensor_mask=acts_nonzero)

    # ! (4B) Use this to compute the loss effect if this feature is ablated
    # which is just the negative of the change in logprobs
    loss_contribution = eindex(-contribution_to_logprobs, correct_tokens, "batch seq [batch seq]")


    # ! (5) Store the results in a SequenceMultiGroupData object

    # Now that we've indexed everything, construct the batch of SequenceData objects
    sequence_groups_data = []
    group_sizes_cumsum = np.cumsum([0] + [len(indices) for indices in indices_dict.values()]).tolist()
    for group_idx, group_name in enumerate(indices_dict.keys()):
        seq_data = [
            SequenceData(
                token_ids = token_ids[i].tolist(),
                feat_acts = [round(f, 4) for f in feat_acts_coloring[i].tolist()],
                loss_contribution = loss_contribution[i].tolist(),
                token_logits = feat_logits[token_ids[i]].tolist(),
                top_token_ids = top_contribution_to_logits.indices[i].tolist(),
                top_logits = top_contribution_to_logits.values[i].tolist(),
                bottom_token_ids = bottom_contribution_to_logits.indices[i].tolist(),
                bottom_logits = bottom_contribution_to_logits.values[i].tolist(),
            )
            for i in range(group_sizes_cumsum[group_idx], group_sizes_cumsum[group_idx+1])
        ]
        sequence_groups_data.append(SequenceGroupData(group_name, seq_data))

    return SequenceMultiGroupData(sequence_groups_data)


@torch.inference_mode()
def parse_prompt_data(
    tokens: Int[Tensor, "batch seq"],
    str_toks: list[str],
    sae_vis_data: SaeVisData,
    feat_acts: Float[Tensor, "seq feats"],
    feature_resid_dir: Float[Tensor, "feats d_model"],
    resid_post: Float[Tensor, "seq d_model"],
    W_U: Float[Tensor, "d_model d_vocab"],
    feature_idx: Optional[list[int]] = None,
    num_top_features: int = 10,
) -> dict[str, tuple[list[int], list[str]]]:
    """
    Gets data needed to create the sequences in the prompt-centric vis (displaying dashboards for the most relevant
    features on a prompt).
    
    This function exists so that prompt dashboards can be generated without using our AutoEncoder or
    TransformerLens(Wrapper) classes. 
    
    Args:
        tokens: Int[Tensor, "batch seq"]
            The tokens we'll be using to get the feature activations. Note that we might not be using all of them; the
            number used is determined by `fvp.total_batch_size`.

        str_toks:  list[str]
            The tokens as a list of strings, so that they can be visualized in HTML.

        sae_vis_data: SaeVisData
             The object storing all data for each feature. We'll set each `feature_data.prompt_data` to the
             data we get from `prompt`.

        feat_acts: Float[Tensor, "seq feats"]
            The activations values of the features across the sequence.

        feature_resid_dir: Float[Tensor, "feats d_model"]
            The directions that each feature writes to the residual stream.

        resid_post: Float[Tensor, "seq d_model"]
            The activations of the final layer of the model before the unembed. 
        
        W_U: Float[Tensor, "d_model d_vocab"]
            The model's unembed weights for the logit lens.

        feature_idx: list[int] or None
            The features we're actually computing. These might just be a subset of the model's full features.
            
        num_top_features: int
            The number of top features to display in this view, for any given metric.

    Returns:
        scores_dict: dict[str, tuple[list[int], list[str]]]
            A dictionary mapping keys like "act_quantile|'django' (0)" to a tuple of lists, where the first list is the
            feature indices, and the second list is the string-formatted values of the scores.

    As well as returning this dictionary, this function will also set `FeatureData.prompt_data` for each feature in
    `sae_vis_data` (this is necessary for getting the prompts in the prompt-centric vis). Note this design choice could
    have been done differently (i.e. have this function return a list of the prompt data for each feature). I chose this
    way because it means the FeatureData._get_html_data_prompt_centric can work fundamentally the same way as 
    FeatureData._get_html_data_feature_centric, rather than treating the prompt data object as a different kind of
    component in the vis.
    """
    if feature_idx is None: 
        feature_idx = list(sae_vis_data.feature_data_dict.keys())
    n_feats = len(feature_idx)
    assert feature_resid_dir.shape[0] == n_feats, f"The number of features in feature_resid_dir ({feature_resid_dir.shape[0]}) does not match the number of feature indices ({n_feats})"

    assert feat_acts.shape[1] == n_feats, f"The number of features in feat_acts ({feat_acts.shape[1]}) does not match the number of feature indices ({n_feats})"

    feats_loss_contribution = torch.empty(size=(n_feats, tokens.shape[1]-1), device=device)
    # Some logit computations which we only need to do once
    # correct_token_unembeddings = model_wrapped.W_U[:, tokens[0, 1:]] # [d_model seq]
    orig_logits = (resid_post / resid_post.std(dim=-1, keepdim=True)) @ W_U # [seq d_vocab]
    raw_logits = feature_resid_dir @ W_U # [feats d_vocab]

    for i, feat in enumerate(feature_idx):

        # ! Calculate the sequence data for each feature, and store it as FeatureData.prompt_data

        # Get this feature's output vector, using an outer product over the feature activations for all tokens
        resid_post_feature_effect = einops.einsum(feat_acts[:, i], feature_resid_dir[i], "seq, d_model -> seq d_model")
        
        # Ablate the output vector from the residual stream, and get logits post-ablation
        new_resid_post = resid_post - resid_post_feature_effect
        new_logits = (new_resid_post / new_resid_post.std(dim=-1, keepdim=True)) @ W_U

        # Get the top5 & bottom5 changes in logits (don't bother with `efficient_topk` cause it's small)
        contribution_to_logprobs = orig_logits.log_softmax(dim=-1) - new_logits.log_softmax(dim=-1)
        top_contribution_to_logits = TopK(contribution_to_logprobs[:-1], k=5)
        bottom_contribution_to_logits = TopK(contribution_to_logprobs[:-1], k=5, largest=False)

        # Get the change in loss (which is negative of change of logprobs for correct token)
        loss_contribution = eindex(-contribution_to_logprobs[:-1], tokens[0, 1:], "seq [seq]")
        feats_loss_contribution[i, :] = loss_contribution

        # Store the sequence data
        sae_vis_data.feature_data_dict[feat].prompt_data = SequenceData(
            token_ids = tokens.squeeze(0).tolist(),
            feat_acts = [round(f, 4) for f in feat_acts[:, i].tolist()],
            loss_contribution = [0.0] + loss_contribution.tolist(),
            token_logits = raw_logits[i, tokens.squeeze(0)].tolist(),
            top_token_ids = top_contribution_to_logits.indices.tolist(),
            top_logits = top_contribution_to_logits.values.tolist(),
            bottom_token_ids = bottom_contribution_to_logits.indices.tolist(),
            bottom_logits = bottom_contribution_to_logits.values.tolist(),
        )

    # ! Lastly, return a dictionary mapping each key like 'act_quantile|"django" (0)' to a list of feature indices & scores

    # Get a dict with keys like f"act_quantile|'My' (1)" and values (feature indices list, feature score values list)
    scores_dict: dict[str, tuple[list[int], list[str]]] = {}

    for seq_pos, seq_key in enumerate([f"{t!r} ({i})" for i, t in enumerate(str_toks)]):

        # Filter the feature activations, since we only need the ones that are non-zero
        feat_acts_nonzero_filter = utils.to_numpy(feat_acts[seq_pos] > 0)
        feat_acts_nonzero_locations = np.nonzero(feat_acts_nonzero_filter)[0].tolist()
        _feat_acts = feat_acts[seq_pos, feat_acts_nonzero_filter] # [feats_filtered,]
        _feature_idx = np.array(feature_idx)[feat_acts_nonzero_filter]

        if feat_acts_nonzero_filter.sum() > 0:
            k = min(num_top_features, _feat_acts.numel())
            
            # Get the top features by activation size. This is just applying a TopK function to the feat acts (which
            # were stored by the code before this). The feat acts are formatted to 3dp.
            act_size_topk = TopK(_feat_acts, k=k, largest=True)
            top_features = _feature_idx[act_size_topk.indices].tolist()
            formatted_scores = [f"{v:.3f}" for v in act_size_topk.values]
            scores_dict[f"act_size|{seq_key}"] = (top_features, formatted_scores)

            # Get the top features by activation quantile. We do this using the `feature_act_quantiles` object, which
            # was stored `sae_vis_data`. This quantiles object has a method to return quantiles for a given set of
            # data, as well as the precision (we make the precision higher for quantiles closer to 100%, because these
            # are usually the quantiles we're interested in, and it lets us to save space in `feature_act_quantiles`).
            act_quantile, act_precision = sae_vis_data.feature_act_quantiles.get_quantile(_feat_acts, feat_acts_nonzero_locations)
            act_quantile_topk = TopK(act_quantile, k=k, largest=True)
            act_formatting = [f".{act_precision[i]-2}%" for i in act_quantile_topk.indices]
            top_features = _feature_idx[act_quantile_topk.indices].tolist()
            formatted_scores = [f"{v:{f}}" for v, f in zip(act_quantile_topk.values, act_formatting)]
            scores_dict[f"act_quantile|{seq_key}"] = (top_features, formatted_scores)

        # We don't measure loss effect on the first token
        if seq_pos == 0:
            continue
        
        # Filter the loss effects, since we only need the ones which have non-zero feature acts on the tokens before them
        prev_feat_acts_nonzero_filter = utils.to_numpy((feat_acts[seq_pos - 1] > 0))
        _loss_contribution = feats_loss_contribution[prev_feat_acts_nonzero_filter, seq_pos-1] # [feats_filtered,]
        _feature_idx_prev = np.array(feature_idx)[prev_feat_acts_nonzero_filter]

        if prev_feat_acts_nonzero_filter.sum() > 0:
            k = min(num_top_features, _loss_contribution.numel())

            # Get the top features by loss effect. This is just applying a TopK function to the loss effects (which were
            # stored by the code before this). The loss effects are formatted to 3dp. We look for the most negative
            # values, i.e. the most loss-reducing features.
            loss_contribution_topk = TopK(_loss_contribution, k=k, largest=False)
            top_features = _feature_idx_prev[loss_contribution_topk.indices].tolist()
            formatted_scores = [f"{v:+.3f}" for v in loss_contribution_topk.values]
            scores_dict[f"loss_effect|{seq_key}"] = (top_features, formatted_scores)
    return scores_dict


@torch.inference_mode()
def get_prompt_data(
    sae_vis_data: SaeVisData,
    prompt: str,
    num_top_features: int,
) -> dict[str, tuple[list[int], list[str]]]:
    '''
    Gets data that will be used to create the sequences in the prompt-centric HTML visualisation, i.e. an object of
    type SequenceData for each of our features.
    
    Args:
        sae_vis_data:     The object storing all data for each feature. We'll set each `feature_data.prompt_data` to the
                          data we get from `prompt`.
        prompt:           The prompt we'll be using to get the feature activations.#
        num_top_features: The number of top features we'll be getting data for.

    Returns:
        scores_dict:      A dictionary mapping keys like "act_quantile|0" to a tuple of lists, where the first list is
                          the feature indices, and the second list is the string-formatted values of the scores.

    As well as returning this dictionary, this function will also set `FeatureData.prompt_data` for each feature in
    `sae_vis_data`. This is because the prompt-centric vis will call `FeatureData._get_html_data_prompt_centric` on each
    feature data object, so it's useful to have all the data in once place! Even if this will get overwritten next
    time we call `get_prompt_data` for this same `sae_vis_data` object.
    '''

    # ! Boring setup code
    feature_idx = list(sae_vis_data.feature_data_dict.keys())
    feature_idx_tensor = torch.tensor(feature_idx, device=device)
    encoder = sae_vis_data.encoder; assert isinstance(encoder, AutoEncoder)
    model = sae_vis_data.model; assert isinstance(model, HookedTransformer)
    cfg = sae_vis_data.cfg; assert isinstance(cfg.hook_point, str), f"{cfg.hook_point=}, expected a string"

    str_toks: list[str] = model.tokenizer.tokenize(prompt) # type: ignore
    tokens = model.tokenizer.encode(prompt, return_tensors="pt").to(device) # type: ignore
    assert isinstance(tokens, torch.Tensor)

    model_wrapped = TransformerLensWrapper(model, cfg.hook_point)

    feature_act_dir = encoder.W_enc[:, feature_idx] # [d_in feats]
    feature_out_dir = encoder.W_dec[feature_idx] # [feats d_in]
    feature_resid_dir = to_resid_dir(feature_out_dir, model_wrapped) # [feats d_model]
    assert feature_act_dir.T.shape == feature_out_dir.shape == (len(feature_idx), encoder.cfg.d_in)

    # ! Define hook functions to cache all the info required for feature ablation, then run those hook fns

    resid_post, act_post = model_wrapped(tokens, return_logits=False)
    resid_post: Tensor = resid_post.squeeze(0)
    feat_acts = compute_feat_acts(act_post, feature_idx_tensor, encoder).squeeze(0) # [seq feats]

    
    # ! Use the data we've collected to make the scores_dict and update the sae_vis_data
    scores_dict = parse_prompt_data(
        tokens = tokens,
        str_toks = str_toks,
        sae_vis_data = sae_vis_data,
        feat_acts = feat_acts,
        feature_resid_dir = feature_resid_dir,
        resid_post = resid_post,
        W_U = model.W_U,
        feature_idx = feature_idx,
        num_top_features = num_top_features,
    )

    return scores_dict




import torch
from torch import Tensor
import pprint
import torch.nn as nn
import torch.nn.functional as F
from dataclasses import dataclass
from jaxtyping import Float, Int, Bool
from transformer_lens import utils, HookedTransformer
from transformer_lens.hook_points import HookPoint
from typing import overload, Tuple, Literal
import re


DTYPES = {"fp32": torch.float32, "fp16": torch.float16, "bf16": torch.bfloat16}

@dataclass
class AutoEncoderConfig:
    '''Class for storing configuration parameters for the autoencoder'''
    seed: int = 42
    batch_size: int = 32
    buffer_mult: int = 384
    epochs: int = 10
    lr: float = 1e-3
    num_tokens: int = int(2e9)
    l1_coeff: float = 3e-4
    beta1: float = 0.9
    beta2: float = 0.99
    dict_mult: int = 8
    seq_len: int = 128
    d_in: int = 2048
    enc_dtype: str = "fp32"
    remove_rare_dir: bool = False
    model_batch_size: int = 64
    device: str = "cuda"

    def __post_init__(self):
        '''Using kwargs, so that we can pass in a dict of parameters which might be
        a superset of the above, without error.'''
        self.buffer_size = self.batch_size * self.buffer_mult
        self.buffer_batches = self.buffer_size // self.seq_len
        self.dtype = DTYPES[self.enc_dtype]
        self.d_hidden = self.d_in * self.dict_mult



class AutoEncoder(nn.Module):
    def __init__(self, cfg: AutoEncoderConfig):
        super().__init__()

        self.cfg = cfg
        torch.manual_seed(cfg.seed)

        # W_enc has shape (d_in, d_encoder), where d_encoder is a multiple of d_in (cause dictionary learning; overcomplete basis)
        self.W_enc = nn.Parameter(torch.nn.init.kaiming_uniform_(torch.empty(cfg.d_in, cfg.d_hidden, dtype=cfg.dtype)))
        self.W_dec = nn.Parameter(torch.nn.init.kaiming_uniform_(torch.empty(cfg.d_hidden, cfg.d_in, dtype=cfg.dtype)))
        self.b_enc = nn.Parameter(torch.zeros(cfg.d_hidden, dtype=cfg.dtype))
        self.b_dec = nn.Parameter(torch.zeros(cfg.d_in, dtype=cfg.dtype))
        self.W_dec.data[:] = self.W_dec / self.W_dec.norm(dim=-1, keepdim=True)

        self.to(cfg.device)

    def forward(self, x: torch.Tensor):
        x_cent = x - self.b_dec
        acts = F.relu(x_cent @ self.W_enc + self.b_enc)
        x_reconstruct = acts @ self.W_dec + self.b_dec
        l2_loss = (x_reconstruct.float() - x.float()).pow(2).sum(-1).mean(0)
        l1_loss = self.cfg.l1_coeff * (acts.float().abs().sum())
        loss = l2_loss + l1_loss
        return loss, x_reconstruct, acts, l2_loss, l1_loss

    @torch.no_grad()
    def remove_parallel_component_of_grads(self):
        W_dec_normed = self.W_dec / self.W_dec.norm(dim=-1, keepdim=True)
        W_dec_grad_proj = (self.W_dec.grad * W_dec_normed).sum(-1, keepdim=True) * W_dec_normed
        self.W_dec.grad -= W_dec_grad_proj

    @classmethod
    def load_from_hf(cls, version):
        """
        Loads the saved autoencoder from HuggingFace. 

        Note, this is a classmethod, because we'll be using it as `auto_encoder = AutoEncoder.load_from_hf("run1")`

        Version is expected to be an int, or "run1" or "run2"

        version 25 is the final checkpoint of the first autoencoder run,
        version 47 is the final checkpoint of the second autoencoder run.
        """

        assert version in ["run1", "run2"]
        version = 25 if version=="run1" else 47

        # Download config
        cfg = utils.download_file_from_hf("NeelNanda/sparse_autoencoder", f"{version}_cfg.json")
        assert isinstance(cfg, dict)
        # Remove unnecessary params, and rename d_mlp to d_in
        cfg.pop("buffer_batches", None)
        cfg.pop("buffer_size", None)
        cfg["d_in"] = cfg.pop("d_mlp")
        # Convert the config to AutoEncoderConfig class, then create the Autoencoder
        cfg = AutoEncoderConfig(**cfg)
        self = cls(cfg=cfg)

        # Load in state dict
        state_dict = utils.download_file_from_hf("NeelNanda/sparse_autoencoder", f"{version}.pt", force_is_torch=True)
        assert isinstance(state_dict, dict)
        self.load_state_dict(state_dict)

        return self

    def __repr__(self) -> str:
        return f"AutoEncoder(d_in={self.cfg.d_in}, dict_mult={self.cfg.dict_mult})"




# # ==============================================================
# # ! TRANSFORMERS
# # This returns the activations & resid_pre as well (optionally)
# # ==============================================================


class TransformerLensWrapper(nn.Module):
    '''
    This class wraps around & extends the TransformerLens model, so that we can make sure things like the forward
    function have a standardized signature.
    '''
    def __init__(self, model: HookedTransformer, hook_point: str):
        super().__init__()
        assert hook_point in model.hook_dict, f"Error: hook_point={hook_point!r} must be in model.hook_dict"
        self.model = model
        self.hook_point = hook_point

        # Get the layer (so we can do the early stopping in our forward pass)
        layer_match = re.match(r'blocks\.(\d+)\.', hook_point)
        assert layer_match, f"Error: expecting hook_point to be 'blocks.{{layer}}.{{...}}', but got {hook_point!r}"
        self.hook_layer = int(layer_match.group(1))

        # Get the hook names for the residual stream (final) and residual stream (immediately after hook_point)
        self.hook_point_resid = utils.get_act_name("resid_post", self.hook_layer)
        self.hook_point_resid_final = utils.get_act_name("resid_post", self.model.cfg.n_layers-1)
        assert self.hook_point_resid in model.hook_dict
        assert self.hook_point_resid_final in model.hook_dict

    @overload
    def forward(
        self,
        tokens: Tensor,
        return_logits: Literal[True],
    ) -> tuple[Tensor, Tensor, Tensor]:
        ...

    @overload
    def forward(
        self,
        tokens: Tensor,
        return_logits: Literal[False],
    ) -> tuple[Tensor, Tensor]:
        ...
    
    def forward(
        self,
        tokens: Int[Tensor, "batch seq"],
        return_logits: bool = True,
    ):
        '''
        Inputs:
            tokens: Int[Tensor, "batch seq"]
                The input tokens, shape (batch, seq)
            return_logits: bool
                If True, returns (logits, residual, activation)
                If False, returns (residual, activation)
        '''

        # Run with hook functions to store the activations & final value of residual stream
        # If return_logits is False, then we compute the last residual stream value but not the logits
        output: Tensor = self.model.run_with_hooks(
            tokens, 
            # stop_at_layer = (None if return_logits else self.hook_layer),
            fwd_hooks = [
                (self.hook_point, self.hook_fn_store_act),
                (self.hook_point_resid_final, self.hook_fn_store_act)
            ]
        )

        # The hook functions work by storing data in model's hook context, so we pop them back out
        activation: Tensor = self.model.hook_dict[self.hook_point].ctx.pop("activation")
        if self.hook_point_resid_final == self.hook_point:
            residual: Tensor = activation
        else:
            residual: Tensor = self.model.hook_dict[self.hook_point_resid_final].ctx.pop("activation")

        if return_logits:
            return output, residual, activation
        return residual, activation

    def hook_fn_store_act(self, activation: torch.Tensor, hook: HookPoint):
        hook.ctx["activation"] = activation

    @property
    def tokenizer(self):
        return self.model.tokenizer

    @property
    def W_U(self):
        return self.model.W_U

    @property
    def W_out(self):
        return self.model.W_out



def to_resid_dir(dir: Float[Tensor, "feats d_in"], model: TransformerLensWrapper):
    """
    Takes a direction (eg. in the post-ReLU MLP activations) and returns the corresponding dir in the residual stream.

    Args:
        dir: 
            The direction in the activations, i.e. shape (feats, d_in) where d_in could be d_model, d_mlp, etc.
        model:
            The model, which should be a HookedTransformerWrapper or similar.
    """
    # If this SAE was trained on the residual stream or attn/mlp out, then we don't need to do anything
    if "resid" in model.hook_point or "_out" in model.hook_point:
        return dir
    
    # If it was trained on the MLP layer, then we apply the W_out map
    elif ("pre" in model.hook_point) or ("post" in model.hook_point):
        return dir @ model.W_out[model.hook_layer]
    
    # Others not yet supported
    else:
        raise NotImplementedError("The hook your SAE was trained on isn't yet supported")



import numpy as np
import itertools
from pathlib import Path
from torch import Tensor
from typing import Optional, Union, Any, Callable
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json
from jaxtyping import Int
import re
from copy import deepcopy
import json

from transformer_lens import HookedTransformer

from sae_vis.model_fns import AutoEncoder

from sae_vis.utils_fns import (
    to_str_tokens,
    QuantileCalculator,
    HistogramData,
    get_decode_html_safe_fn,
    unprocess_str_tok,
    max_or_1,
)
from sae_vis.html_fns import (
    bgColorMap,
    uColorMap,
    HTML,
)
from sae_vis.data_config_classes import (
    SaeVisLayoutConfig,
    SaeVisConfig,
    FeatureTablesConfig,
    ActsHistogramConfig,
    LogitsTableConfig,
    LogitsHistogramConfig,
    SequencesConfig,
    PromptConfig,
    GenericComponentConfig,
)

METRIC_TITLES = {
    "act_size": "Activation Size",
    "act_quantile": "Activation Quantile",
    "loss_effect": "Loss Effect",
}
PRECISION = 4

@dataclass_json
@dataclass
class FeatureTablesData:
    '''
    This contains all the data necessary to make the left-hand tables in prompt-centric visualization. See diagram
    in readme: 

        https://github.com/callummcdougall/sae_vis#data_storing_fnspy

    Inputs:
        neuron_alignment...
            The data for the neuron alignment table (each of its 3 columns). In other words, the data containing which
            neurons in the transformer the encoder feature is most aligned with.

        correlated_neurons...
            The data for the correlated neurons table (each of its 3 columns). In other words, the data containing which
            neurons in the transformer are most correlated with the encoder feature.

        correlated_features...
            The data for the correlated features table (each of its 3 columns). In other words, the data containing
            which features in encoder-B are most correlated with those in the original encoder. Note, this one might be
            absent if we're not using a B-encoder.
    '''
    neuron_alignment_indices: list[int] = field(default_factory=list)
    neuron_alignment_values: list[float] = field(default_factory=list)
    neuron_alignment_l1: list[float] = field(default_factory=list)
    correlated_neurons_indices: list[int] = field(default_factory=list)
    correlated_neurons_pearson: list[float] = field(default_factory=list)
    correlated_neurons_cossim: list[float] = field(default_factory=list)
    correlated_features_indices: list[int] = field(default_factory=list)
    correlated_features_pearson: list[float] = field(default_factory=list)
    correlated_features_cossim: list[float] = field(default_factory=list)

    def _get_html_data(
        self,
        cfg: FeatureTablesConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {},
    ) -> HTML:
        '''
        Returns the HTML for the left-hand tables, wrapped in a 'grid-column' div.

        Note, we only ever use this obj in the context of the left-hand column of the feature-centric vis, and it's
        always the same width & height, which is why there's no customization available for this function.
        '''
        # Crop the lists to `cfg.n_rows` (first checking the config doesn't ask for more rows than we have)
        assert cfg.n_rows <= len(self.neuron_alignment_indices)
        neuron_alignment_indices = self.neuron_alignment_indices[:cfg.n_rows]
        neuron_alignment_values = self.neuron_alignment_values[:cfg.n_rows]
        neuron_alignment_l1 = self.neuron_alignment_l1[:cfg.n_rows]
        correlated_neurons_indices = self.correlated_neurons_indices[:cfg.n_rows]
        correlated_neurons_pearson = self.correlated_neurons_pearson[:cfg.n_rows]
        correlated_neurons_cossim = self.correlated_neurons_cossim[:cfg.n_rows]

        # Read HTML from file, and replace placeholders with real ID values
        html_str = (Path(__file__).parent / "html" / "feature_tables_template.html").read_text()
        html_str = html_str.replace("FEATURE_TABLES_ID", f"feature-tables-{id_suffix}")

        # Add the neuron alignment and neuron correlation data to the object which will be turned into JavaScript
        data: dict[str, Any] = {
            "neuronAlignment": [
                {"index": I, "value": f"{V:+.3f}", "percentageL1": f"{L:.1%}"}
                for I, V, L in zip(neuron_alignment_indices, neuron_alignment_values, neuron_alignment_l1)
            ],
            "correlatedNeurons": [
                {"index": I, "value": f"{P:+.3f}", "percentageL1": f"{C:+.3f}"}
                for I, P, C in zip(correlated_neurons_indices, correlated_neurons_pearson, correlated_neurons_cossim)
            ],
        }

        # If we have correlated features from encoder_B, add that to the JavaScript data
        if len(self.correlated_features_indices) > 0:
            correlated_features_indices = self.correlated_features_indices[:cfg.n_rows]
            correlated_features_pearson = self.correlated_features_pearson[:cfg.n_rows]
            correlated_features_cossim = self.correlated_features_cossim[:cfg.n_rows]

            assert (correlated_features_pearson is not None) and (correlated_features_cossim is not None), "All or none"
            data["correlatedFeaturesB"] = [
                {"index": I, "value": f"{P:+.3f}", "percentageL1": f"{C:+.3f}"}
                for I, P, C in zip(correlated_features_indices, correlated_features_pearson, correlated_features_cossim)
            ]
        # If not, remove that table from the HTML string
        else:
            html_str = re.sub(r'<h4>CORRELATED FEATURES \(B-ENCODER\)</h4>.*?</table>', "", html_str, flags=re.DOTALL)

        return HTML(
            html_data = {column: html_str},
            js_data = {"featureTablesData": {id_suffix: data}}
        )


@dataclass_json
@dataclass
class ActsHistogramData(HistogramData):
    def _get_html_data(
        self,
        cfg: ActsHistogramConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {},
    ) -> HTML:
        '''
        Converts data -> HTML object, for the feature activations histogram (i.e. the histogram over all sampled tokens,
        showing the distribution of activations for this feature).
        '''
        # We can't post-hoc change the number of bins, so check this wasn't changed in the config
        # assert cfg.n_bins == len(self.bar_heights),\
        #     "Can't post-hoc change `n_bins` in histogram config - you need to regenerate data."
        
        # Read HTML from file, and replace placeholders with real ID values
        html_str = (Path(__file__).parent / "html" / "acts_histogram_template.html").read_text()
        html_str = html_str.replace("HISTOGRAM_ACTS_ID", f"histogram-acts-{id_suffix}")

        # Process colors for frequency histogram; it's darker at higher values
        bar_values_normed = [(0.4 * max(self.bar_values) + 0.6 * v) / max(self.bar_values) for v in self.bar_values]
        bar_colors = [bgColorMap(v) for v in bar_values_normed]
        
        # Next we create the data dict
        data: dict[str, Any] = {
            "y": self.bar_heights,
            "x": self.bar_values,
            "ticks": self.tick_vals,
            "colors": bar_colors,
        }
        if self.title is not None:
            data["title"] = self.title

        return HTML(
            html_data = {column: html_str},
            js_data = {"actsHistogramData": {id_suffix: data}},
        )


@dataclass_json
@dataclass
class LogitsHistogramData(HistogramData):
    def _get_html_data(
        self,
        cfg: LogitsHistogramConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {},
    ) -> HTML:
        '''
        Converts data -> HTML object, for the logits histogram (i.e. the histogram over all tokens in the vocab, showing
        the distribution of direct logit effect on that token).
        '''
        # We can't post-hoc change the number of bins, so check this wasn't changed in the config
        # assert cfg.n_bins == len(self.bar_heights),\
        #     "Can't post-hoc change `n_bins` in histogram config - you need to regenerate data."

        # Read HTML from file, and replace placeholders with real ID values
        html_str = (Path(__file__).parent / "html" / "logits_histogram_template.html").read_text()
        html_str = html_str.replace("HISTOGRAM_LOGITS_ID", f"histogram-logits-{id_suffix}")

        data: dict[str, Any] = {
            "y": self.bar_heights,
            "x": self.bar_values,
            "ticks": self.tick_vals,
        }
        if self.title is not None:
            data["title"] = self.title

        return HTML(
            html_data = {column: html_str},
            js_data = {"logitsHistogramData": {id_suffix: data}},
        )


@dataclass_json
@dataclass
class LogitsTableData:

    bottom_token_ids: list[int] = field(default_factory=list)
    bottom_logits: list[float] = field(default_factory=list)
    top_token_ids: list[int] = field(default_factory=list)
    top_logits: list[float] = field(default_factory=list)

    def _get_html_data(
        self,
        cfg: LogitsTableConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {},
    ) -> HTML:
        '''
        Converts data -> HTML object, for the logits table (i.e. the top and bottom affected tokens by this feature).
        '''
        # Crop the lists to `cfg.n_rows` (first checking the config doesn't ask for more rows than we have)
        assert cfg.n_rows <= len(self.bottom_logits)
        bottom_token_ids = self.bottom_token_ids[:cfg.n_rows]
        bottom_logits = self.bottom_logits[:cfg.n_rows]
        top_token_ids = self.top_token_ids[:cfg.n_rows]
        top_logits = self.top_logits[:cfg.n_rows]
        
        # Get the negative and positive background values (darkest when equals max abs)
        max_value = max(max(top_logits[:cfg.n_rows]), -min(bottom_logits[:cfg.n_rows]))
        neg_bg_values = np.absolute(bottom_logits[:cfg.n_rows]) / max_value
        pos_bg_values = np.absolute(top_logits[:cfg.n_rows]) / max_value

        # Get the string tokens, using the decode function
        neg_str = to_str_tokens(decode_fn, bottom_token_ids[:cfg.n_rows])
        pos_str = to_str_tokens(decode_fn, top_token_ids[:cfg.n_rows])

        # Read HTML from file, and replace placeholders with real ID values
        html_str = (Path(__file__).parent / "html" / "logits_table_template.html").read_text()
        html_str = html_str.replace("LOGITS_TABLE_ID", f"logits-table-{id_suffix}")

        # Create object for storing JS data
        data: dict[str, list] = {"negLogits": [], "posLogits": []}

        # Get data for the tables of pos/neg logits
        for i in range(len(neg_str)):
            data["negLogits"].append({
                "symbol": unprocess_str_tok(neg_str[i]),
                "value": round(top_logits[i], 2),
                "color": f"rgba(255,{int(255*(1-neg_bg_values[i]))},{int(255*(1-neg_bg_values[i]))},0.5)"
            })
            data["posLogits"].append({
                "symbol": unprocess_str_tok(pos_str[i]),
                "value": round(top_logits[i], 2),
                "color": f"rgba({int(255*(1-pos_bg_values[i]))},{int(255*(1-pos_bg_values[i]))},255,0.5)"})
            
        return HTML(
            html_data = {column: html_str},
            js_data = {"logitsTableData": {id_suffix: data}},
        )





@dataclass_json
@dataclass
class SequenceData:
    '''
    This contains all the data necessary to make a sequence of tokens in the vis. See diagram in readme: 

        https://github.com/callummcdougall/sae_vis#data_storing_fnspy

    Always-visible data:
        token_ids:              List of token IDs in the sequence
        feat_acts:              Sizes of activations on this sequence
        loss_contribution:   Effect on loss of this feature, for this particular token (neg = helpful)

    Data which is visible on hover:
        token_logits:       The logits of the particular token in that sequence (used for line on logits histogram)
        top_token_ids:     List of the top 5 logit-boosted tokens by this feature
        top_logits:        List of the corresponding 5 changes in logits for those tokens
        bottom_token_ids:  List of the bottom 5 logit-boosted tokens by this feature
        bottom_logits:     List of the corresponding 5 changes in logits for those tokens
    '''
    token_ids: list[int] = field(default_factory=list)
    feat_acts: list[float] = field(default_factory=list)
    loss_contribution: list[float] = field(default_factory=list)

    token_logits: list[float] = field(default_factory=list)
    top_token_ids: list[list[int]] = field(default_factory=list)
    top_logits: list[list[float]] = field(default_factory=list)
    bottom_token_ids: list[list[int]] = field(default_factory=list)
    bottom_logits: list[list[float]] = field(default_factory=list)

    def __post_init__(self) -> None:
        '''
        Filters the logits & token IDs by removing any elements which are zero (this saves space in the eventual
        JavaScript).
        '''
        self.seq_len = len(self.token_ids)
        self.top_logits, self.top_token_ids = self._filter(self.top_logits, self.top_token_ids)
        self.bottom_logits, self.bottom_token_ids = self._filter(self.bottom_logits, self.bottom_token_ids)

    def _filter(
        self,
        float_list: list[list[float]],
        int_list: list[list[int]]
    ) -> tuple[list[list[float]], list[list[int]]]:
        '''
        Filters the list of floats and ints, by removing any elements which are zero. Note - the absolute values of the
        floats are monotonic non-increasing, so we can assume that all the elements we keep will be the first elements
        of their respective lists. Also reduces precisions of feature activations & logits.
        '''
        # Next, filter out zero-elements and reduce precision
        float_list = [[round(f, PRECISION) for f in floats if abs(f) > 1e-6] for floats in float_list]
        int_list = [ints[:len(floats)] for ints, floats in zip(int_list, float_list)]
        return float_list, int_list
    

    def _get_html_data(
        self,
        cfg: PromptConfig | SequencesConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {},
    ) -> HTML:
        '''
        Args:

        Returns:
            js_data: list[dict[str, Any]]
                The data for this sequence, in the form of a list of dicts for each token (where the dict stores things
                like token, feature activations, etc).
        '''
        assert isinstance(cfg, (PromptConfig, SequencesConfig)), f"Invalid config type: {type(cfg)}"
        seq_group_id = component_specific_kwargs.get("seq_group_id", None)
        max_act_color = component_specific_kwargs.get("max_act_color", None)
        max_loss_color = component_specific_kwargs.get("max_loss_color", None)
        bold_idx = component_specific_kwargs.get("bold_idx", None)
        permanent_line = component_specific_kwargs.get("permanent_line", False)
        first_in_group = component_specific_kwargs.get("first_in_group", True)
        title = component_specific_kwargs.get("title", None)
        hover_above = component_specific_kwargs.get("hover_above", False)

        # If we didn't supply a sequence group ID, then we assume this sequence is on its own, and give it a unique ID
        if seq_group_id is None: seq_group_id = f"prompt-{column:03d}"
        
        # If we didn't specify bold_idx, then set it to be the midpoint
        if bold_idx is None: bold_idx = self.seq_len // 2

        # If we only have data for the bold token, we pad out everything with zeros or empty lists
        only_bold = isinstance(cfg, SequencesConfig) and not(cfg.compute_buffer)
        if only_bold:
            feat_acts = [self.feat_acts[0] if i == bold_idx else 0.0 for i in range(self.seq_len)]
            loss_contribution = [self.loss_contribution[0] if i == bold_idx + 1 else 0.0 for i in range(self.seq_len)]
            pos_ids = [self.top_token_ids[0] if i == bold_idx + 1 else [] for i in range(self.seq_len)]
            neg_ids = [self.bottom_token_ids[0] if i == bold_idx + 1 else [] for i in range(self.seq_len)]
            pos_val = [self.top_logits[0] if i == bold_idx + 1 else [] for i in range(self.seq_len)]
            neg_val = [self.bottom_logits[0] if i == bold_idx + 1 else [] for i in range(self.seq_len)]
        else:
            feat_acts = deepcopy(self.feat_acts)
            loss_contribution = deepcopy(self.loss_contribution)
            pos_ids = deepcopy(self.top_token_ids)
            neg_ids = deepcopy(self.bottom_token_ids)
            pos_val = deepcopy(self.top_logits)
            neg_val = deepcopy(self.bottom_logits)

        # Get values for converting into colors later
        bg_denom = max_or_1(self.feat_acts) if (max_act_color is None) else max_act_color
        u_denom = max_or_1(self.loss_contribution, abs=True) if (max_loss_color is None) else max_loss_color
        bg_values = (np.maximum(feat_acts, 0.0) / max(1e-4, bg_denom)).tolist()
        u_values = (np.array(loss_contribution) / max(1e-4, u_denom)).tolist()

        # If we sent in a prompt rather than this being sliced from a longer sequence, then the pos_ids etc will be shorter
        # than the token list by 1, so we need to pad it at the first token
        if isinstance(cfg, PromptConfig):
            assert len(pos_ids) == len(neg_ids) == len(pos_val) == len(neg_val) == len(self.token_ids) - 1,\
                "If this is a single prompt, these lists must be the same length as token_ids or 1 less"
            pos_ids = [[]] + pos_ids
            neg_ids = [[]] + neg_ids
            pos_val = [[]] + pos_val
            neg_val = [[]] + neg_val
        assert len(pos_ids) == len(neg_ids) == len(pos_val) == len(neg_val) == len(self.token_ids),\
            "If this is part of a sequence group etc are given, they must be the same length as token_ids"

        # Process the tokens to get str toks
        toks = to_str_tokens(decode_fn, self.token_ids)
        pos_toks = [to_str_tokens(decode_fn, pos) for pos in pos_ids]
        neg_toks = [to_str_tokens(decode_fn, neg) for neg in neg_ids]

        # Define the JavaScript object which will be used to populate the HTML string
        js_data_list = []

        for i in range(len(self.token_ids)):

            # We might store a bunch of different case-specific data in the JavaScript object for each token. This is
            # done in the form of a disjoint union over different dictionaries (which can each be empty or not), this
            # minimizes the size of the overall JavaScript object. See function in `tokens_script.js` for more.
            kwargs_bold: dict[str, bool] = {}
            kwargs_hide: dict[str, bool] = {}
            kwargs_this_token_active: dict[str, Any] = {}
            kwargs_prev_token_active: dict[str, Any] = {}
            kwargs_hover_above: dict[str, bool] = {}


            # Get args if this is the bolded token (we make it bold, and maybe add permanent line to histograms)
            if bold_idx is not None and bold_idx == i:
                kwargs_bold["isBold"] = True
                if permanent_line: kwargs_bold["permanentLine"] = True

            # If we only have data for the bold token, we hide all other tokens' hoverdata (and skip other kwargs)
            if only_bold and isinstance(bold_idx, int) and (i not in {bold_idx, bold_idx + 1}):
                kwargs_hide["hide"] = True

            else:
                # Get args if we're making the tooltip hover above token (default is below)
                if hover_above:
                    kwargs_hover_above["hoverAbove"] = True

                # If feature active on this token, get background color and feature act (for hist line)
                if abs(feat_acts[i]) > 1e-8:
                    kwargs_this_token_active = dict(
                        featAct = round(feat_acts[i], PRECISION),
                        bgColor = bgColorMap(bg_values[i]),
                    )

                # If prev token active, get the top/bottom logits table, underline color, and loss effect (for hist line)
                pos_toks_i, neg_toks_i, pos_val_i, neg_val_i = pos_toks[i], neg_toks[i], pos_val[i], neg_val[i]
                if len(pos_toks_i) + len(neg_toks_i) > 0:
                    # Create dictionary
                    kwargs_prev_token_active = dict(
                        posToks = pos_toks_i,
                        negToks = neg_toks_i,
                        posVal = pos_val_i,
                        negVal = neg_val_i,
                        lossEffect = round(loss_contribution[i], PRECISION),
                        uColor = uColorMap(u_values[i]),
                    )

            js_data_list.append(dict(
                tok = unprocess_str_tok(toks[i]),
                tokID = self.token_ids[i],
                tokenLogit = round(self.token_logits[i], PRECISION),
                **kwargs_bold,
                **kwargs_this_token_active,
                **kwargs_prev_token_active,
                **kwargs_hover_above,
            ))

        # Create HTML string (empty by default since sequences are added by JavaScript) and JS data    
        html_str = ""
        js_seq_group_data: dict[str, Any] = {"data": [js_data_list]}

        # Add group-specific stuff if this is the first sequence in the group
        if first_in_group:

            # Read HTML from file, replace placeholders with real ID values
            html_str = (Path(__file__).parent / "html" / "sequences_group_template.html").read_text()
            html_str = html_str.replace("SEQUENCE_GROUP_ID", seq_group_id)

            # Get title of sequence group, and the idSuffix to match up with a histogram
            js_seq_group_data["idSuffix"] = id_suffix
            if title is not None: js_seq_group_data["title"] = title

        return HTML(
            html_data = {column: html_str},
            js_data = {"tokenData": {seq_group_id: js_seq_group_data}}
        )


@dataclass_json
@dataclass
class SequenceGroupData:
    '''
    This contains all the data necessary to make a single group of sequences (e.g. a quantile in prompt-centric
    visualization). See diagram in readme: 

        https://github.com/callummcdougall/sae_vis#data_storing_fnspy

    Inputs:
        title:      The title that this sequence group will have, if any. This is used in `_get_html_data`. The titles
                    will actually be in the HTML strings, not in the JavaScript data.
        seq_data:   The data for the sequences in this group.
    '''
    title: str = ""
    seq_data: list[SequenceData] = field(default_factory=list)
    
    def __len__(self) -> int:
        return len(self.seq_data)
    
    def _get_html_data(
        self,
        cfg: SequencesConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {}

        # These default values should be correct when we only have one sequence group, because when we call this from
        # a SequenceMultiGroupData we'll override them)
    ) -> HTML:
        '''
        This creates a single group of sequences, i.e. title plus some number of vertically stacked sequences.

        Note, `column` is treated specially here, because it might overflow (could be a tuple).

        Args:
            decode_fn:      Mapping from token IDs to string tokens.
            bold_idx:       If supplied, then we bold the token at this index. Default is the middle token. Note, this
                            argument is only allowed if the sequence group contains just one sequence.
            group_size:     Max size of sequences in the group (i.e. we truncate after this many, if argument supplied).
            max_act_color:  If supplied, then we use this as the most extreme value (for coloring by feature act).
            permanent_line: If True, the bolded token for first seq in group will have a permanent line on histograms.
            seq_group_id:   The id of the sequence group div. This will usually be passed as e.g. "seq-group-001".
            column:         The index of this column. Note that it can be a tuple, if our SequenceMultiGroupData object
                            has overflowed into a new column.
        
        Returns:
            html_obj:       Object containing the HTML and JavaScript data for this seq group.
        '''
        seq_group_id = component_specific_kwargs.get("seq_group_id", None)
        group_size = component_specific_kwargs.get("group_size", None)
        max_act_color = component_specific_kwargs.get("max_act_color", None)
        max_loss_color = component_specific_kwargs.get("max_loss_color", None)

        # Get the data that will go into the div (list of list of dicts, i.e. containing all data for seqs in group). We
        # start with the title.
        html_obj = HTML()

        # If seq_group_id is not supplied, then we assume this is the only sequence in the column, and we name the group
        # after the column
        if seq_group_id is None: seq_group_id = f"seq-group-{column:03d}"

        # If max_act_color not supplied, use the max over all sequences in this group. Same for loss color
        if max_act_color is None: max_act_color = max_or_1([act for seq in self.seq_data for act in seq.feat_acts])
        if max_loss_color is None: max_loss_color = max_or_1([loss for seq in self.seq_data for loss in seq.loss_contribution], abs=True)
        
        # Accumulate the HTML data for each sequence in this group
        for i, seq in enumerate(self.seq_data[:group_size]):
            html_obj += seq._get_html_data(
                cfg = cfg,
                # pass in a PromptConfig object
                decode_fn = decode_fn,
                id_suffix = id_suffix,
                column = column,
                component_specific_kwargs = dict(
                    bold_idx = cfg.buffer[0],
                    permanent_line = False, # in a group, we're never showing a permanent line (only for single seqs)
                    max_act_color = max_act_color,
                    max_loss_color = max_loss_color,
                    seq_group_id = seq_group_id,
                    first_in_group = (i == 0),
                    title = self.title,
                )
            )
        
        return html_obj



@dataclass_json
@dataclass
class SequenceMultiGroupData:
    '''
    This contains all the data necessary to make multiple groups of sequences (e.g. the different quantiles in the
    prompt-centric visualization). See diagram in readme: 

        https://github.com/callummcdougall/sae_vis#data_storing_fnspy
    '''
    seq_group_data: list[SequenceGroupData] = field(default_factory=list)

    def __getitem__(self, idx: int) -> SequenceGroupData:
        return self.seq_group_data[idx]

    def _get_html_data(
        self,
        cfg: SequencesConfig,
        decode_fn: Callable[[int | list[int]], str | list[str]],
        id_suffix: str,
        column: int | tuple[int, int],
        component_specific_kwargs: dict[str, Any] = {},
    ) -> HTML:
        '''
        Args:

        Returns:
            html_obj:  Object containing the HTML and JavaScript data for these multiple seq groups.
        '''
        assert isinstance(column, int)

        # Get max activation value & max loss contributions, over all sequences in all groups
        max_act_color = max_or_1([
            act for group in self.seq_group_data for seq in group.seq_data for act in seq.feat_acts
        ])
        max_loss_color = max_or_1([
            loss for group in self.seq_group_data for seq in group.seq_data for loss in seq.loss_contribution
        ], abs=True)

        # Get the correct column indices for the sequence groups, depending on how group_wrap is configured. Note, we 
        # deal with overflowing columns by extending the dictionary, i.e. our column argument isn't just `column`, but
        # is a tuple of `(column, x)` where `x` is the number of times we've overflowed. For instance, if we have mode
        # 'stack-none' then our columns are `(column, 0), (column, 1), (column, 1), (column, 1), (column, 2), ...`
        n_groups = len(self.seq_group_data)
        n_quantile_groups = n_groups - 1
        match cfg.stack_mode:
            case "stack-all":
                # Here, we stack all groups into 1st column
                cols = [column for i in range(n_groups)]
            case "stack-quantiles":
                # Here, we give 1st group its own column, and stack all groups into second column
                cols = [(column, 0)] + [(column, 1) for _ in range(n_quantile_groups)]
            case "stack-none":
                # Here, we stack groups into columns as [1, 3, 3, ...]
                cols = [(column, 0), *[(column, 1 + int(i/3)) for i in range(n_quantile_groups)]]
            case _:
                raise ValueError(f"Invalid stack_mode: {cfg.stack_mode}. Expected in 'stack-x' for x='all', 'quantiles', 'none'")

        # Create the HTML object, and add all the sequence groups to it, possibly across different columns
        html_obj = HTML()
        for i, (col, group_size, sequences_group) in enumerate(zip(cols, cfg.group_sizes, self.seq_group_data)):
            html_obj += sequences_group._get_html_data(
                cfg = cfg,
                decode_fn = decode_fn,
                id_suffix = id_suffix,
                column = col,
                component_specific_kwargs = dict(
                    group_size = group_size,
                    max_act_color = max_act_color,
                    max_loss_color = max_loss_color,
                    seq_group_id = f"seq-group-{column}-{i}", # we label our sequence groups with (index, column)
                )
            )

        return html_obj



GenericData = Union[
    FeatureTablesData,
    ActsHistogramData,
    LogitsTableData,
    LogitsHistogramData,
    SequenceMultiGroupData,
    SequenceData,
]


@dataclass
class FeatureData:
    '''
    This contains all the data necessary to make the feature-centric visualization, for a single feature. See
    diagram in readme: 

        https://github.com/callummcdougall/sae_vis#data_storing_fnspy

    Args:
        feature_idx:    Index of the feature in question (not used within this class's methods, but used elsewhere).
        cfg:            Contains layout parameters which are important in the `get_html` function.
        
        The other args are the 6 possible components we might have in the feature-centric vis, i.e. this is where we
        store the actual data. Note that one of these arguments is `prompt_data` which is only applicable in the prompt-
        centric view.

    This is used in both the feature-centric and prompt-centric views. In the feature-centric view, a single one
    of these objects creates the HTML for a single feature (i.e. a full screen). In the prompt-centric view, a single
    one of these objects will create one column of the full screen vis.
    '''
    feature_tables_data: FeatureTablesData = field(default_factory = lambda: FeatureTablesData())
    acts_histogram_data: ActsHistogramData = field(default_factory = lambda: ActsHistogramData())
    logits_table_data: LogitsTableData = field(default_factory = lambda: LogitsTableData())
    logits_histogram_data: LogitsHistogramData = field(default_factory = lambda: LogitsHistogramData())
    sequence_data: SequenceMultiGroupData = field(default_factory = lambda: SequenceMultiGroupData())
    prompt_data: SequenceData = field(default_factory = lambda: SequenceData())

    def get_component_from_config(self, config: GenericComponentConfig) -> GenericData:
        '''
        Given a config object, returns the corresponding data object stored by this instance. For instance, if the input
        is an `FeatureTablesConfig` instance, then this function returns `self.feature_tables_data`.
        '''
        CONFIG_CLASS_MAP = {
            FeatureTablesConfig.__name__: self.feature_tables_data,
            ActsHistogramConfig.__name__: self.acts_histogram_data,
            LogitsTableConfig.__name__: self.logits_table_data,
            LogitsHistogramConfig.__name__: self.logits_histogram_data,
            SequencesConfig.__name__: self.sequence_data,
            PromptConfig.__name__: self.prompt_data,
        }
        config_class_name = config.__class__.__name__
        assert config_class_name in CONFIG_CLASS_MAP, f"Invalid component config: {config_class_name}"
        return CONFIG_CLASS_MAP[config_class_name]
    
    def _get_html_data_feature_centric(
        self,
        layout_cfg: SaeVisLayoutConfig,
        decode_fn: Callable,
    ) -> HTML:
        '''
        Returns the HTML object for a single feature-centric view. These are assembled together into the full feature-
        centric view.

        Args:
            decode_fn:  We use this function to decode the token IDs into string tokens. 

        Returns:
            html_obj.html_data:
                Contains a dictionary with keys equal to columns, and values equal to the HTML strings. These will be
                turned into grid-column elements, and concatenated.
            html_obj.js_data:
                Contains a dictionary with keys = component names, and values = JavaScript data that will be used by the
                scripts we'll eventually dump in.
        '''
        # Create object to store all HTML
        html_obj = HTML()

        # For every column in this feature-centric layout, we add all the components in that column
        for column_idx, column_components in layout_cfg.columns.items():

            for component_config in column_components:

                component = self.get_component_from_config(component_config)
                
                html_obj += component._get_html_data(
                    cfg = component_config,
                    decode_fn = decode_fn,
                    column = column_idx,
                    id_suffix = "0", # we only use this if we have >1 set of histograms, i.e. prompt-centric vis
                )

        return html_obj

    def _get_html_data_prompt_centric(
        self,
        layout_cfg: SaeVisLayoutConfig,
        decode_fn: Callable,
        column_idx: int,
        bold_idx: int,
        title: str,
    ) -> HTML:
        '''
        Returns the HTML object for a single column of the prompt-centric view. These are assembled together into a full
        screen of a prompt-centric view, and then they're further assembled together into the full prompt-centric view.

        Args:
            decode_fn:  We use this function to decode the token IDs into string tokens. 
            column_idx: This method only gives us a single column (of the prompt-centric vis), so we need to know which
                        column this is (for the JavaScript data).
            title:      The title for this column, which will be used in the JavaScript data.

        Returns:
            html_obj.html_data:
                Contains a dictionary with the single key `str(column_idx)`, representing the single column. This will
                become a single grid-column element, and will get concatenated with others of these.
            html_obj.js_data:
                Contains a dictionary with keys = component names, and values = JavaScript data that will be used by the
                scripts we'll eventually dump in.
        '''
        # Create object to store all HTML
        html_obj = HTML()

        # Verify that we only have a single column
        assert layout_cfg.columns.keys() == {0},\
            f"prompt_centric_layout should only have 1 column, instead found cols {layout_cfg.columns.keys()}"
        assert layout_cfg.prompt_cfg is not None,\
            "prompt_centric_cfg should include a PromptConfig, but found None"
        if layout_cfg.seq_cfg is not None:
            assert (layout_cfg.seq_cfg.n_quantiles == 0) or (layout_cfg.seq_cfg.stack_mode == "stack-all"),\
            "prompt_centric_layout should have stack_mode='stack-all' if n_quantiles > 0, so that it fits in 1 col"

        # For every component in the single column of this prompt-centric layout, add all the components in that column
        for component_config in layout_cfg.columns[0]:

            component = self.get_component_from_config(component_config)

            html_obj += component._get_html_data(
                cfg = component_config,
                decode_fn = decode_fn,
                column = column_idx,
                id_suffix = str(column_idx),
                component_specific_kwargs = dict(  # only used for the SequenceData (the prompt)
                    bold_idx = bold_idx,
                    permanent_line = True,
                    hover_above = True,
                ),
            )
        
        # Add the title in JavaScript, and the empty title element in HTML
        html_obj.html_data[column_idx] = f"<div id='column-{column_idx}-title'></div>\n{html_obj.html_data[column_idx]}"
        html_obj.js_data["gridColumnTitlesData"] = {str(column_idx): title}

        return html_obj


@dataclass_json
@dataclass
class _SaeVisData:
    '''
    Dataclass which is used to store the data for the SaeVisData class. It excludes everything which isn't easily
    serializable, only saving the raw data.
    '''
    feature_data_dict: dict[int, FeatureData] = field(default_factory=dict)
    feature_act_quantiles: QuantileCalculator = field(default_factory=QuantileCalculator)

    @classmethod
    def from_dict(cls, data: dict) -> '_SaeVisData':
        ... # just for type hinting; the method comes from 'dataclass_json'

    def to_dict(self) -> dict:
        ... # just for type hinting; the method comes from 'dataclass_json'




@dataclass
class SaeVisData:
    '''
    This contains all the data necessary for constructing the feature-centric visualization, over multiple 
    features (i.e. being able to navigate through them). See diagram in readme:

        https://github.com/callummcdougall/sae_vis#data_storing_fnspy

    Args: 
        feature_data_dict:      Contains the data for each individual feature-centric vis.
        feature_act_quantiles:  Contains the quantiles of activation values for each feature (used for rank-ordering
                                features in the prompt-centric vis).
        cfg:                    The vis config, used for the both the data gathering and the vis layout.
        model:                  The model which our encoder was trained on.
        encoder:                The encoder used to get the feature activations.
        encoder_B:              The encoder used to get the feature activations for the second model (if applicable).
    '''
    feature_data_dict: dict[int, FeatureData] = field(default_factory=dict)
    feature_act_quantiles: QuantileCalculator = field(default_factory=QuantileCalculator)
    cfg: SaeVisConfig = field(default_factory=SaeVisConfig)

    model: Optional[HookedTransformer] = None
    encoder: Optional[AutoEncoder] = None
    encoder_B: Optional[AutoEncoder] = None

    def update(self, other: "SaeVisData") -> None:
        '''
        Updates a SaeVisData object with the data from another SaeVisData object. This is useful during the
        `get_feature_data` function, since this function is broken up into different groups of features then merged
        together.
        '''
        if other is None: return
        self.feature_data_dict.update(other.feature_data_dict)
        self.feature_act_quantiles.update(other.feature_act_quantiles)


    @classmethod
    def create(
        cls,
        encoder: AutoEncoder,
        model: HookedTransformer,
        tokens: Int[Tensor, "batch seq"],
        cfg: SaeVisConfig,
        encoder_B: Optional[AutoEncoder] = None,
    ) -> "SaeVisData":
        
        from sae_vis.data_fetching_fns import get_feature_data

        sae_vis_data = get_feature_data(
            encoder = encoder,
            model = model,
            tokens = tokens,
            cfg = cfg,
            encoder_B = encoder_B,
        )
        sae_vis_data.cfg = cfg
        sae_vis_data.model = model
        sae_vis_data.encoder = encoder
        sae_vis_data.encoder_B = encoder_B

        return sae_vis_data


    def save_feature_centric_vis(
       self,
       filename: str | Path,
       feature_idx: Optional[int] = None,
    ) -> None:
        '''
        Returns the HTML string for the view which lets you navigate between different features.

        Args:
            model:          Used to get the tokenizer (for converting token IDs to string tokens).
            filename:       The HTML filepath we'll save the visualization to.
            feature_idx:    This is the default feature index we'll start on. If None, we use the first feature.
        '''
        # Initialize the object we'll eventually get_html from
        HTML_OBJ = HTML()

        # Set the default argument for the dropdown (i.e. when the page first loads)
        first_feature = next(iter(self.feature_data_dict)) if (feature_idx is None) else feature_idx

        # Get tokenize function (we only need to define it once)
        assert self.model is not None
        decode_fn = get_decode_html_safe_fn(self.model.tokenizer)

        # For each FeatureData object, we get the html_obj for it's feature-centric vis, and merge it with HTML_OBJ
        for feature, feature_data in self.feature_data_dict.items():

            # Get the HTML object for a single-feature view
            html_obj = feature_data._get_html_data_feature_centric(self.cfg.feature_centric_layout, decode_fn)

            # Add the JavaScript, and arbitrarily set html_str to be the first feature's html_str (they're all same!)
            HTML_OBJ.js_data[str(feature)] = deepcopy(html_obj.js_data)
            if feature == first_feature: HTML_OBJ.html_data = deepcopy(html_obj.html_data)
        
        # Save our full HTML
        HTML_OBJ.get_html(
            layout = self.cfg.feature_centric_layout,
            filename = filename,
            first_key = str(first_feature),
        )


    def save_prompt_centric_vis(
        self,
        prompt: str,
        filename: str | Path,
        metric: Optional[str] = None,
        seq_pos: Optional[int] = None,
        num_top_features: int = 10,
    ):
        '''
        Returns the HTML string for the view which lets you navigate between different features.

        Args:
            prompt:     The user-input prompt.
            model:      Used to get the tokenizer (for converting token IDs to string tokens).
            filename:   The HTML filepath we'll save the visualization to.
            metric:     This is the default scoring metric we'll start on. If None, we use 'act_quantile'.
            seq_pos:    This is the default seq pos we'll start on. If None, we use 0.
        '''
        # Initialize the object we'll eventually get_html from
        HTML_OBJ = HTML()

        # Run forward passes on our prompt, and store the data within each FeatureData object as `self.prompt_data` as
        # well as returning the scores_dict (which maps from score hash to a list of feature indices & formatted scores)
        from sae_vis.data_fetching_fns import get_prompt_data
        scores_dict = get_prompt_data(
            sae_vis_data = self,
            prompt = prompt,
            num_top_features = num_top_features,
        )

        # Get all possible values for dropdowns
        str_toks = self.model.tokenizer.tokenize(prompt) # type: ignore
        str_toks = [t.replace("|", "│") for t in str_toks] # vertical line -> pipe (hacky, so key splitting on | works)
        str_toks_list = [f"{t!r} ({i})" for i, t in enumerate(str_toks)]
        metric_list = ["act_quantile", "act_size", "loss_effect"]

        # Get default values for dropdowns
        first_metric = "act_quantile" or metric
        first_seq_pos = str_toks_list[0 if seq_pos is None else seq_pos]
        first_key = f"{first_metric}|{first_seq_pos}"

        # Get tokenize function (we only need to define it once)
        assert self.model is not None
        decode_fn = get_decode_html_safe_fn(self.model.tokenizer)

        # For each (metric, seqpos) object, we merge the prompt-centric views of each of the top features, then we merge
        # these all together into our HTML_OBJ
        for (_metric, _seq_pos) in itertools.product(metric_list, range(len(str_toks))):

            # Create the key for this given combination of metric & seqpos, and get our top features & scores
            key = f"{_metric}|{str_toks_list[_seq_pos]}"
            if key not in scores_dict:
                continue
            feature_idx_list, scores_formatted = scores_dict[key]
            
            # Create HTML object, to store each feature column for all the top features for this particular key
            html_obj = HTML()

            for i, (feature_idx, score_formatted) in enumerate(zip(feature_idx_list, scores_formatted)):
                
                # Get HTML object at this column (which includes JavaScript to dynamically set the title)
                html_obj += self.feature_data_dict[feature_idx]._get_html_data_prompt_centric(
                    layout_cfg = self.cfg.prompt_centric_layout,
                    decode_fn = decode_fn,
                    column_idx = i,
                    bold_idx = _seq_pos,
                    title = f"<h3>#{feature_idx}<br>{METRIC_TITLES[_metric]} = {score_formatted}</h3><hr>",
                )

            # Add the JavaScript (which includes the titles for each column)
            HTML_OBJ.js_data[key] = deepcopy(html_obj.js_data)

            # Set the HTML data to be the one with the most columns (since different options might have fewer cols)
            if len(HTML_OBJ.html_data) < len(html_obj.html_data):
                HTML_OBJ.html_data = deepcopy(html_obj.html_data)
        
        # Check our first key is in the scores_dict (if not, we should pick a different key)
        assert first_key in scores_dict, f"Key {first_key} not found in {scores_dict.keys()=}. Have you tried \
computing your initial data with more features and/or tokens, to make sure you have enough positive examples?"

        # Save our full HTML
        HTML_OBJ.get_html(
            layout = self.cfg.prompt_centric_layout,
            filename = filename,
            first_key = first_key,
        )


    def save_json(self: "SaeVisData", filename: str | Path) -> None:
        '''
        Saves an SaeVisData instance to a JSON file. The config, model & encoder arguments must be user-supplied.
        '''
        if isinstance(filename, str): filename = Path(filename)
        assert filename.suffix == ".json", "Filename must have a .json extension"

        _self = _SaeVisData(
            feature_data_dict = self.feature_data_dict,
            feature_act_quantiles = self.feature_act_quantiles,
        )

        with open(filename, "w") as f:
            json.dump(_self.to_dict(), f)


    @classmethod
    def load_json(
        cls,
        filename: str | Path,
        cfg: SaeVisConfig,
        model: HookedTransformer,
        encoder: AutoEncoder,
        encoder_B: AutoEncoder,
    ) -> "SaeVisData":
        '''
        Loads an SaeVisData instance from JSON file. The config, model & encoder arguments must be user-supplied.
        '''
        if isinstance(filename, str): filename = Path(filename)
        assert filename.suffix == ".json", "Filename must have a .json extension"

        with open(filename, "r") as f:
            data = json.load(f)

        _self = _SaeVisData.from_dict(data)

        self = SaeVisData(
            feature_data_dict = _self.feature_data_dict,
            feature_act_quantiles = _self.feature_act_quantiles,
            cfg = cfg,
            model = model,
            encoder = encoder,
            encoder_B = encoder_B,
        )

        return self



from setuptools import setup, find_packages

setup(
    name = 'sae-vis',
    version = '0.2.6',
    packages = find_packages(),
    install_requires = [
        'torch',
        'einops',
        'datasets',
        'dataclasses-json',
        'jaxtyping',
        # 'eindex',
    ],
    # dependency_links = [
    #     'git+https://github.com/callummcdougall/eindex.git#egg=eindex'  # Add the link with egg specification
    # ],
    include_package_data = True,
    author = 'Callum McDougall',
    author_email = 'cal.s.mcdougall@gmail.com',
    description = "Open-source SAE visualizer, based on Anthropic's published visualizer.",
    url = 'https://github.com/callummcdougall/sae_visualizer',
    classifiers = [
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
)